from datetime import datetime

from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.project import Project
from mage_ai.data_preparation.models.project.constants import FeatureUUID
from mage_ai.kernels.models import KernelProcess, KernelWrapper
from mage_ai.kernels.utils import find_ipykernel_launchers_info_async
from mage_ai.orchestration.db import safe_db_query
from mage_ai.server.active_kernel import (
    interrupt_kernel,
    restart_kernel,
    start_kernel,
    switch_active_kernel,
)
from mage_ai.server.kernels import DEFAULT_KERNEL_NAME, KernelName, kernel_managers
from mage_ai.services.ssh.aws.emr.utils import tunnel
from mage_ai.shared.singletons.memory import get_memory_manager_controller


class KernelResource(GenericResource):
    @classmethod
    @safe_db_query
    async def collection(self, query, meta, user, **kwargs):
        kernels = []

        if Project().is_feature_enabled(FeatureUUID.AUTOMATIC_KERNEL_CLEANUP):
            # Only do this every minute
            if int(datetime.utcnow().timestamp()) % 60 == 0:
                kill_count, memory_freed = KernelProcess.terminate_inactive(
                    await find_ipykernel_launchers_info_async(),
                )
                if kill_count >= 1:
                    print(
                        f'[KernelResource] Automatic kernel cleanup: {kill_count} '
                        f'kernels terminated, {(memory_freed / 1024**3):.3f} GBs if memory freed.'
                    )

        for kernel_name in KernelName:
            kernel = kernel_managers[kernel_name]
            if kernel.has_kernel:
                kernels.append(KernelWrapper(kernel))

        return self.build_result_set(
            kernels,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    def member(self, pk, user, **kwargs):
        kernel_fallback = None
        kernels_by_id = {}

        for kernel_name in KernelName:
            kernel = kernel_managers[kernel_name]
            if kernel.has_kernel:
                kernels_by_id[kernel.kernel_id] = kernel

                if not kernel_fallback:
                    kernel_fallback = kernel

        kernel = kernels_by_id.get(pk)
        if not kernel:
            kernel = kernel_fallback
        if not kernel:
            kernel = kernel_managers[DEFAULT_KERNEL_NAME]

        return self(KernelWrapper(kernel), user, **kwargs)

    @safe_db_query
    def update(self, payload, **kwargs):
        action_type = payload.get('action_type')

        switch_active_kernel(self.model.kernel_name)

        if 'interrupt' == action_type:
            get_memory_manager_controller().stop_all_events()
            interrupt_kernel()
        elif 'restart' == action_type:
            try:
                get_memory_manager_controller().stop_all_events()
                restart_kernel()
            except RuntimeError as e:
                # RuntimeError: Cannot restart the kernel. No previous call to 'start_kernel'.
                if 'start_kernel' in str(e):
                    get_memory_manager_controller().stop_all_events()
                    start_kernel()

        def _callback(*args, **kwargs):
            tunnel(
                kernel_name=self.model.kernel_name,
                reconnect=True,
                validate_conditions=True,
            )

        self.on_update_callback = _callback

        return self
